structure ListAdaptive : LIST_ADAPTIVE =
struct

datatype 'a u = NIL | CONS of ('a  * 'a u AdaptiveMeta.Box.t)
type 'a t = 'a u AdaptiveMeta.Box.t
type 'a alist = 'a t

(** Shorthands for adaptivity primitives **)
val put = Adaptive.put
val putTh = Adaptive.putTh
val mkPut = Adaptive.mkPut 
val get = Adaptive.get
val new = Adaptive.new
val deref = Adaptive.deref
val change = Adaptive.change

(**********************************************************************
 ** lengthLessThan
 **********************************************************************)
fun lengthLessThan_static n l = 
   let
      fun len (m, l) =
         if m >= n
            then false
         else 
            case l of
               nil => true
             | _::t => len  (m + 1, t)
   in
     len (0, l)
   end

fun lengthLessThan n = afn l =>
   let
      mfun len (m, l) =
         if m >= n
            then false
         else 
            case get $ l of
               NIL => true
             | CONS (_, t) => len $ (m + 1, t)
   in
     putTh $ (afn () => len $ (0, l))
   end

(**********************************************************************
 ** reverse
 **********************************************************************)
(** Fill this in **)
afun reverse l = l

(**********************************************************************
 ** filter
 **********************************************************************)
fun filter f = afn l =>
   let
      val putM = mkPut $ ()
      mfun filt l = 
         case get $ l of
            NIL => NIL
          | CONS (h, t) => 
	    if f h then
	      CONS (h, putM $ (h, filt $ t))
	    else
	      filt $ t
   in
      putTh $ (afn () => filt $ l)
   end

(**********************************************************************
 ** map
 **********************************************************************)
fun map_static f l = 
   let
      fun map l = 
         case l of
            nil => nil
          | h::t =>  (f h)::(map t)
   in
      map l
   end

fun map f = afn l =>
   let
      val putM = mkPut $ ()
      mfun mmap l = 
         case get $ l of
            NIL => NIL
          | CONS (h, tl) =>  CONS (f h, putM $ (h, mmap $ tl))
   in
      putTh $ (afn () => mmap $ l)
   end

(**********************************************************************
 ** partition
 **********************************************************************)
(** Fill this in **)
fun partition p = afn l => 
    let
      val putMA  = mkPut $ ()
      val putMB = mkPut $ ()
    in
      (l,l)
    end

(**********************************************************************
 ** reduce
 **********************************************************************)
(** Fill this in **)
fun reduce f = afn l => put $ 0

fun reduce f = afn l =>
  let 
    mfun m (l,r) = 
      case get $ l of
	NIL => r
      | CONS(h,t) => m $ (t, f (r,h))
  in
    putTh $ (afn () => 
	     case get $ l of
	       NIL => 0
	     | CONS(h,t) => m $ (t,h))
  end

fun reduce_static f l = 
  case l of
    nil => 0
  | h::t => List.foldr f 0 l

(**********************************************************************
 ** Merge 
 **********************************************************************)
fun merge lt = afn (a : 'a t, b : 'a t) =>
    let
      val putM = mkPut $ ()
       mfun mmerge (a, b) =
         case (get $ a, get $ b) of
           (NIL,b) => b
         | (a,NIL) => a
         | (CONS(ha,ta),CONS(hb,tb)) => 
           if lt (ha, hb)
           then CONS (ha, putM $ (ha, mmerge $ (ta,b)))
           else CONS (hb, putM $ (hb, mmerge $ (a,tb)))

    in
      putTh $ (afn () => mmerge $ (a, b))
    end

(**********************************************************************
 ** Merge Sort
 **********************************************************************)
structure HashF = HashFamily(val M = 2)
mfun split l = 
    let 
      val hashFun = HashF.hashFun ()
      fun randFun x = hashFun (Adaptive.hash x) = 0
    in 
      partition randFun $ l
    end

fun msort lt = afn (l : 'a t) =>
    let
      mfun ms l = putTh $ (afn () => 
        if get $ (lengthLessThan 2 $ l)
        then get $ l
        else 
	  let
            val (even, odd) = split $ l
            val even_s = ms $ even
	    val odd_s = ms $ odd
	    val l_s = merge lt $ (even_s, odd_s)
          in
            get $ l_s
          end)
    in
      ms $ (l)
    end

(**********************************************************************
 ** Input Generator
 **********************************************************************)
val SEED = 512894
fun generateRandomList size =
  let
    val state = Random.rand (134567,SEED)
    val randInt = fn () => Random.randInt state
    fun rl(n',i) =
      if n'=0 then nil
      else (randInt (),i)::(rl (n'-1,i+1))
    val l = ListMergeSort.sort (fn ((x,_),(y,_)) => x < y) (rl (size,0))
    val l = List.map (fn (x,y) => y) l
  in
    l
  end

fun generateSortedList size =
  let
    fun rl i =
      if i=size-1 then nil
      else i::(rl (i+1))
  in
    rl 0
  end

(**********************************************************************
 ** Mutator
 **********************************************************************)
fun mutator 
    testName
    (inputToString: 'a t -> string) 
    (resultToString: 'b -> string) 
    (resultToString_static: 'c -> string)
    (f: 'a t -$> 'b) 
    (f_static: 'a list -> 'c)
    (compareResults: 'b * 'c -> bool)
    (l: 'a list) : unit = 
let
  fun toList (l: 'a t) : 'a list = 
    case (Adaptive.deref l) of 
      NIL => nil 
    |  CONS(h,t) => h::(toList t) 
  
  fun fromList (l: 'a list) : 'a t =  
    case l of 
      nil => new NIL 
    | h::t => new (CONS(h, fromList t) )
  
  fun checkAndPrintResult (l,r,r_static) =  
    case !r of 
      Result.Exn e => raise e 
    | Result.Value v => 
      let 
	val i = inputToString l
	val rs = resultToString v
(*
	val _ = print ("Check = " ^ Bool.toString (compareResults (v,r_static)) ^ "\n")
*)
      in 
	(print ("Input = " ^ i ^ "\n"); 
	 print ("Result = " ^  rs ^ "\n"))
      end
  
   fun delete cl =  
    case (Adaptive.deref cl) of
      NIL => (NONE) 
    | CONS(h,t) => (change (cl, deref t); SOME h) 
  
   fun insert h cl =  
    case h of 
      NONE => () 
    |  SOME v => (change (cl,(CONS (v,new (deref cl)))))
  
   fun GC () = (MLton.GC.collect ();
		print ("Bytes live: " ^
                       (IntInf.toString (MLton.GC.Statistics.lastBytesLive ())) ^ "\n"))

   fun perform str f  = 
     let
       val _ = print ("---Performing: " ^ str ^ "---\n")
       val (r,stats) = Stats.wrap (f, ())
       val _ = print ("Run Time for " ^ str ^ ":\n" ^ 
		      (Stats.toString stats) ^ "\n")
     in
       r
     end

   fun deleteAndPropagate cl = 
     let 
       val h = delete cl
     in
       (Adaptive.propagate (); h)
     end
     
   fun insertAndPropagate h cl = 
     let 
       val _ = insert h cl
     in
       Adaptive.propagate ()
     end


   fun repeat f m = 
     if m = 0 then 
       f ()
     else
       (f (); repeat f (m-1))
     
   val _ = AdaptiveMeta.init ()
   val cl = fromList l
   val _ = GC ()

   val _ = print ("--- START TEST [ " ^ testName ^ "] \n")

   val r_static = perform "Static From-Scratch" 
			  (fn () => f_static l)
   val r = perform "Self-Adjusting From-Scratch" 
		   (fn () => Adaptive.call (f, cl))

   val _ = checkAndPrintResult (cl,r,r_static)

   val h = perform "Delete and Propagate" (fn () => deleteAndPropagate cl)
   val _ = checkAndPrintResult (cl,r,r_static)

   val _ = perform "Insert and Propagate" (fn () => insertAndPropagate h cl)
   val _ = checkAndPrintResult (cl,r,r_static)
   val _ = GC ()
   val _ = print ("--- END TEST [" ^ testName ^  "] \n")
in 
  () 
end

fun aListToString toString l = 
  let
    val max = 10
    fun f (l,i) = 
      if (i > max) then "..."
      else
	case (deref l) of 
	  NIL => "nil"
	| CONS(h,t) => (toString h) ^ "::" ^ (f (t,i+1))
  in
    f (l,0)
  end

fun listToString toString l = 
  let
    val max = 10
    fun f (l,i) = 
      if (i > max) then "..."
      else
	case (l) of 
	  nil => "nil"
	| h::t => (toString h) ^ "::" ^ (f (t,i+1))
  in
    f (l,0)
  end


val intAListToString  = aListToString Int.toString
val intListToString  = listToString Int.toString
fun testToString ts l = (ts l) 

fun compareListResults (r,r_s) = 
  case (Adaptive.deref r, r_s) of
    (NIL,nil) => true
  | (CONS(h,t),h_s::t_s) => (h_s=h) andalso (compareListResults (t,t_s))
  | _ => false

fun compareEqResults (r,r_s) = 
  (Adaptive.deref r = r_s)

fun testLengthLessThan (n,l) = 
  mutator "lengthLessThan"
	  (testToString  intAListToString)
	  (testToString  (Bool.toString o deref))
          (testToString  (Bool.toString))
	  (lengthLessThan n) 
	  (lengthLessThan_static n)
	  compareEqResults
	  l 

fun testMap f l = 
  mutator "map"
	  (testToString intAListToString)
          (testToString intAListToString)
          (testToString intListToString)
          (map f)
	  (map_static f)
	  compareListResults
          l 

fun testPartition f l = 
  let
    afun partOne l = 
      let val (a,b) = partition f $ l
      in a
      end

    afun partTwo l = 
      let val (a,b) = partition f $ l
      in b
      end

    fun partOne_static l   = 
      let val (a,b) = List.partition f  l
      in a
      end

    fun partTwo l = 
      let val (a,b) = List.partition f  l
      in b
      end

  in
    (
    mutator "partitionTrue"
	    (testToString intAListToString)
	    (testToString intAListToString)
	    (testToString intListToString)
            partOne 
	    partOne_static
	    compareListResults
            l 
(*  ; 
    mutator "partitionFalse"
	    (testToString intAListToString)
	    (testToString intAListToString)
            partTwo
            l 
*)
    )
  end

fun testFilter pred l = 
  mutator "filter" 
	  (testToString intAListToString)
          (testToString intAListToString)
	  (testToString intListToString)
          (filter pred)
	  (List.filter pred)
	  compareListResults
	  l

fun testReverse l = 
  mutator "reverse" 
	  (testToString intAListToString)
          (testToString intAListToString)
	  (testToString intListToString)
          reverse
	  List.rev
	  compareListResults
	  l

fun testReduce f l = 
  mutator "reduce" 
	  (testToString intAListToString)
	  (testToString  (Int.toString o deref))
          (testToString  (Int.toString))
          (reduce f)
          (reduce_static f)
	  compareEqResults
	  l

fun testMSort lessThan l = 
  mutator "msort"
	  (testToString intAListToString)
          (testToString intAListToString)
	  (testToString intListToString)
          (msort lessThan)
	  (ListMergeSort.sort (not o lessThan))
	  compareListResults
	  l
 
end
