structure QuickHullAdaptive : sig include APPLICATION val quick_hull: it -$> ot end =
struct
   structure Poly = Polynomial
   structure Geom = GeometryTwoD
   structure Point = Geom.Point
   val SamplingRatio = 4
   structure HashF = HashFamily(val M = SamplingRatio)

   type it = Point.t ListAdaptive.t
   type ot = Point.t ListAdaptive.t

   structure List = ListAdaptive
   open Adaptive
   open List

   val mkRandSplit : unit -> Point.t -> bool = fn () =>
      let val hashFun = HashF.hashFun ()
      in fn (pt : Point.t) => hashFun (Word.toIntX (Hash.default pt)) = 0
      end

   fun select f (a,b) = if f (a, b) then a else b

   val combine = List.combine mkRandSplit

   afun split (bp1, bp2, l, hull) =
      let
         val {putM = putML, ...} = PutMemoTable.mkPut $ ()

         fun above (pl,pr) p =
            (Poly.eval0sgn o Geom.lineSideTest)
            ((Geom.newLine (pl, pr)), p)

         mfun splitM (p1, p2, l, hull) =
            if get $ (List.lengthLessThan 1 $ l)
               then putML $ (p1, CONS (p1, hull))
            else
               let
                  fun distanceToLine p = 
                     (Poly.eval0 o Geom.distanceToLine) 
                     (p, (p1,p2))
                  val selectMax = select (fn (p1,p2) => 
                     distanceToLine p1 > distanceToLine p2)
                  val max = get $ (combine selectMax $ l)
                  val (left,right) = List.filter2 (above (p1,max), above (max,p2)) $ l
               in
                  splitM $ (p1, max, left, splitM $ (max, p2, right, hull))
               end
         val (p1, p2) = (get $ bp1, get $ bp2)
      in
         splitM $ (p1, p2, List.filter (above (p1, p2)) $ l, hull)
      end

   afun quick_hull (l : it) : ot =
      if get $ (List.lengthLessThan 2 $ l)
         then l
      else 
         let
            val selectMin = select (Poly.eval0sgn o Geom.toLeft)
            val selectMax = select (Poly.eval0sgn o Geom.toRight)
            val min = combine selectMin $ l
            val max = combine selectMax $ l

            val lower = split $ (max, min, l, put $ NIL)
            val hull = split $ (min, max, l, lower) 
         in
            hull
         end

   val quick_hull : it -$> ot = quick_hull
   fun run l = Adaptive.call (quick_hull, l)
end
