(**********************************************************************
 ** adaptive_meta.sig
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The signature for the meta-operations of the adaptive library
 **********************************************************************)

signature ADAPTIVE_META =
sig
   structure Box :
      sig
         type 'a t
         val eq : 'a t Eq.t
         val hash : 'a t Hash.t
         val key : 'a t Key.t
         val toString : ('a -> string) -> 'a t -> string
      end

   (* Meta *)
   val init : unit -> unit

   val newEq : 'a Eq.t * 'a -> 'a Box.t
   val new : 'a -> 'a Box.t
   val change : 'a Box.t * 'a -> unit
   val deref : 'a Box.t -> 'a
   val hasNoReads: 'a Box.t -> bool

   type ('a, 'b) call_fn
   val call : ('a, 'b) call_fn * 'a -> 'b Result.t ref
   val propagate : unit -> int * bool

   val stats : unit -> string list
end
