(**********************************************************************
 ** adaptive.sig
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The signature for the adaptive library
 **********************************************************************)

signature ADAPTIVE =
sig
   include ADAPTIVE_META
   where type ('a, 'b) call_fn = 'a -$> 'b

   val get : 'a Box.t -$> 'a

   type ('a,'ax) putX = 'ax -$> 'a Box.t
   type 'a putTh = ('a, unit -$> 'a) putX
   type 'a put = ('a, 'a) putX

   val putTh : 'a putTh
   val put : 'a put
   val putThEq : 'a Eq.t -> 'a putTh
   val putEq : 'a Eq.t -> 'a put

   structure PutMemoTable :
      sig
         type ('key,'a,'ax) putMX = 'key * 'ax -$> 'a Box.t
         type ('key,'a) putMTh = ('key, 'a, unit -$> 'a) putMX
         type ('key,'a) putM = ('key, 'a, 'a) putMX
          
         (* putMTh: ---A---th()---W---
          * putM: ---AW---
          * mkPutM: ---A---...---WWW---
          *)
         type ('key,'a) t = {putMTh: ('key,'a) putMTh,
                             putM: ('key,'a) putM,
                             mkPutM: unit -$> ('key,'a) putM}
         val mkPut : unit -$> ('key,'a) t
         val mkPutEq : 'key Key.t * 'a Eq.t -$> ('key,'a) t
      end

   val mkPut : unit -$> ('key * 'a -$> 'a Box.t)

   val memoContEq : ('res Eq.t * ('arg -$> 'res)) -> 
                  ('arg -$> 'res)
   val memoFixEq : 'arg Key.t * 'res Eq.t * (('arg -$> 'res) -> ('arg -$> 'res)) ->
                 ('arg -$> 'res)
   val mkMemoFixEq : 'arg Key.t * 'res Eq.t * (('arg -$> 'res) -> ('arg -$> 'res)) ->
                   (unit -$> ('arg -$> 'res))

   val memoEq : 'arg Key.t * 'res Eq.t * ('arg -$> 'res) -> ('arg -$> 'res)
   val mkMemoEq : 'arg Key.t * 'res Eq.t * ('arg -$> 'res) -> (unit -$> ('arg -$> 'res))

   val memoCont : ('arg -$> 'res) -> ('arg -$> 'res)
   val memoFix : (('arg -$> 'res) -> ('arg -$> 'res)) -> ('arg -$> 'res)
   val mkMemoFix : (('arg -$> 'res) -> ('arg -$> 'res)) -> (unit -$> ('arg -$> 'res))
   val memo : ('arg -$> 'res) -> ('arg -$> 'res)
   val mkMemo : ('arg -$> 'res) -> (unit -$> ('arg -$> 'res))
   val hash: 'a -> int

end
