(**********************************************************************
 ** eq.sml
 **
 ** Copyright 2008 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of equality functions
 **********************************************************************)

structure Eq :> EQ =
struct
   type 'a t = 'a * 'a -> bool
   val default = MLton.equal
   val falsee = fn (_,_) => false
   val pointer = MLton.eq

   val bool: bool t = default
   val exn: exn t = default
   val int: int t = default
   val list = fn eqA => fn (al1,al2) =>
      ListPair.allEq eqA (al1,al2)
   val option = fn eqA => fn (ao1,ao2) =>
      (case (ao1,ao2) of
          (NONE, NONE) => true
        | (SOME a1, SOME a2) => eqA (a1, a2)
        | _ => false)
   val real: real t = default
   val string: string t = default
   val sum = fn (eqA,eqB) => fn (abs1,abs2) =>
      (case (abs1,abs2) of
          (Sum.INL a1, Sum.INL a2) => eqA (a1, a2)
        | (Sum.INR b1, Sum.INR b2) => eqB (b1, b2)
        | _ => false)
   val tuple2 = fn (eqA,eqB) => fn ((a1,b1),(a2,b2)) => 
      eqA (a1,a2) andalso eqB (b1,b2)
   val tuple3 = fn (eqA,eqB,eqC) => fn ((a1,b1,c1),(a2,b2,c2)) => 
      eqA (a1,a2) andalso eqB (b1,b2) andalso eqC (c1,c2)
   val tuple4 = fn (eqA,eqB,eqC,eqD) => fn ((a1,b1,c1,d1),(a2,b2,c2,d2)) => 
      eqA (a1,a2) andalso eqB (b1,b2) andalso eqC (c1,c2) andalso eqD (d1,d2)
   val unit = fn ((),()) => true
   val word: word t = default
end
