(**********************************************************************
 ** leftist-heap.sml
 **
 ** Copyright 2008 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** Written by Umut A. Acar and Guy Blelloch
 ** Modified by Matthew Fluet
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: An implementation of leftist heaps
 ** This code mostly copied from Purely Functional Data Structures
 ** (p. 20), by Chris Okasaki.
 **********************************************************************)

functor BuildLeftistHeap(structure Elt: PRIORITY_QUEUE_ELT) :> 
   PRIORITY_QUEUE where type Elt.t = Elt.t =
struct
   structure Elt = Elt

   datatype t = E | T of int * Elt.t * t * t
      
   fun rank tr =
      case tr of
         E => 0
       | T (r, _, _, _) => r
   fun makeT (x, a, b) =
      if rank a >= rank b then
         T (rank b + 1, x, a, b)
      else
         T (rank a + 1, x, b, a)

   val empty : t = E

   fun isEmpty tr =
      case tr of
         E => true
       | _ => false

   fun merge (tr1, tr2) =
      let
         fun merge (tr1, tr2) =
            case (tr1, tr2) of
               (_, E) => tr1
             | (E, _) => tr2
             | (T (_, x, a1, b1), T (_, y, a2, b2)) =>
                  (case Elt.cmp (x, y) of
                      LESS => makeT (x, a1, merge (b1, tr2))
                    | _ => makeT (y, a2, merge (tr1, b2)))
      in
         merge (tr1, tr2)
      end

   fun insert (tr, x) =
      let
         fun insert tr =
            case tr of
               E => T (1, x, E, E)
             | T (r, y, a, b) =>
                  (case Elt.cmp (x, y) of
                      LESS => T (r + 1, x, tr, E)
                    | _ => makeT (y, a, insert b))
      in
         insert tr
      end
   fun findMin tr =
      case tr of
         E => NONE
       | T (_, x, _, _) => SOME x
   fun deleteMin tr =
      case tr of
         E => NONE
       | T (_, x, a, b) =>
            SOME (merge (a, b), x)
end
