(**********************************************************************
 ** nonadaptive_cps.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of the adaptive library
 **********************************************************************)

structure NonAdaptiveCPS :> NONADAPTIVE_CPS =
struct
   (* Meta *)
   structure Box =
      struct
         type 'a t = 'a
         val eq = fn _ => false
         val hash = fn _ => 0wx0
         val key = {eq = eq, hash = hash}
         fun toString vToString b = vToString b
      end

   exception NonAdaptive
   val new = fn x => x
   val newEq = fn (_, x) => x
   val change = fn _ => raise NonAdaptive
   val deref = fn b => b

   val init = fn _ => raise NonAdaptive
   val propagate = fn _ => raise NonAdaptive
   val stats = fn _ => []

   (* CPS *)
   type 'a cont = 'a -> unit 

   structure Dest =
      struct
         type 'a t = unit
      end

   val alloc = fn (_: 'a Eq.t, k: 'a Dest.t cont) => k ()
   val write = fn (_: 'a Dest.t, x : 'a, k : 'a Box.t cont) => k x
   val read = fn (b : 'a Box.t, k : 'a cont) => k b
   val hasNoReads = fn (_: 'a Box.t) => true

   structure AllocWriteMemoTable =
      struct
         structure Dest =
            struct
               type 'a t = unit
            end
         structure Table =
            struct
               type ('key, 'a) t = unit
               val make = fn _ => ()
               val insert = fn (_,_) => ()
               val _ = insert
            end

         val alloc = fn (_: 'a Eq.t, k: 'a Dest.t cont) => k ()
         val write = fn (_,_) =>
            fn (_: 'a Dest.t, x : 'a, k : 'a Box.t cont) =>
            k x

         val allocWrite = fn (_,_) =>
            fn (_, x, k) =>
            k x
      end

   structure FunMemoTable =
      struct
         val mkFunOne = fn (_: 'arg Key.t) => fn f => fn (arg,k,h) =>
            f (arg,(k,h))
         val mkFunTwo = fn (_: 'arg Key.t) => fn f => fn (arg,k,h) =>
            f (arg,(k,h))
         val mkFun = fn (_: 'arg Key.t) => fn f => fn arg =>
            f arg
      end
   structure ContMemoTable =
      struct
         val mkCont = fn (_: 'a Eq.t) =>
            fn k: 'a cont => k
      end

   (* CPS-Meta *)
   exception InitialRun
   type ('a, 'b) call_fn = ('a * 'b cont) cont
   val call = fn (f : ('a, 'b) call_fn, arg : 'a) =>
      let
         val r = ref (Result.Exn InitialRun)
         fun finish res = r := res
         val k0 = fn (v : 'b) => finish (Result.Value v)
         val () = f (arg, k0)
      in
         r
      end
end

structure NonAdaptiveMeta : NONADAPTIVE_META = NonAdaptiveCPS
