(**********************************************************************
 ** unauto-coerce.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of coercions between (implicit) 
 ** functions and (explicit) continuation-passing-style functions;
 ** the coercions are between the external types only.
 **********************************************************************)

structure UnAutoCoerce : UNAUTO_COERCE =
struct
   type 'a cont = 'a -> unit 
   type ('a, 'b) xfn = ('b cont * exn cont * 'a) cont

   datatype 'a t = Value of 'a | Exn of exn

   val fn_to_xfn : ('a -> 'b) -> ('a, 'b) xfn =
      fn (f : 'a -> 'b) => fn (k : 'b cont, h : exn cont, arg : 'a) =>
      case (Value (f arg) handle exn => Exn exn) of
         Value v => k v
       | Exn e => h e
   exception UnAutoCoerce
   val xfn_to_fn : ('a, 'b) xfn -> ('a -> 'b) =
      fn (xf : ('b cont * exn cont * 'a) -> unit) => fn (arg : 'a) =>
      let
         val r = ref (Exn UnAutoCoerce)
         fun finish res = r := res
         val k0 = fn (v : 'b) => finish (Value v)
         val h0 = fn (e : exn) => finish (Exn e)
         val () = xf (k0, h0, arg)
      in
         case !r of
            Value v => v
          | Exn e => raise e
      end
end
