structure AListMetaAdaptive : ALIST_META_ADAPTIVE =
struct

   datatype u = datatype AListAdaptive.u
   datatype t = datatype AListAdaptive.t
   datatype alist = datatype AListAdaptive.alist

   fun new  v = AdaptiveMeta.new (v)
   fun newEq (eq, v) = AdaptiveMeta.newEq (AListEqAdaptive.eqCell eq, v)
   fun change (l, v) = AdaptiveMeta.change (l, v)
   fun deref l = AdaptiveMeta.deref l

   fun insertAt (l, e) =
      let
         val nl = new (deref l)
         val () = change (l, CONS(e, nl))
      in
         nl
      end
   fun deleteAt l =
      case deref l of
         NIL => ()
       | CONS(_,t) => change (l, deref t)
   fun deleteFirstAt (l, p) =
      case deref l of
         NIL => ()
       | CONS(h,t) =>
            if p h
               then change (l, deref t)
            else deleteFirstAt (l, p)
   fun drop (l, i) =
      if i = 0
         then l
      else (case deref l of
               NIL => l
             | CONS(_,t) => drop (t, i - 1))

   fun length l =
      let
         fun len (l, n) =
            case deref l of
               NIL => n
             | CONS(_,t) => len (t, n+1)
      in
         len (l, 0)
      end

   fun fromList l =
      let
         fun fl l =
            case l of
               nil => new (NIL)
             | h::t => new (CONS (h, fl t))
      in
         fl l
      end
   fun fromListWithLast l =
      let
         fun fl l =
            case l of
               nil => let val l = new NIL
                      in (l, l)
                      end
             | h::nil => let val l = new NIL
                             val l = new (CONS (h, l))
                         in (l,l)
                         end
             | h::t => let val (l,last) = fl t
                       in (new (CONS (h, l)), last)
                       end
      in
         fl l
      end

   fun toList m = 
      case deref m of
         NIL => nil
       | CONS(h,t) => h::(toList t)

   fun toString f m = 
      concat ["[", String.concatWith "," (map f (toList m)), "]"]
end
