structure ListEqAdaptive : LIST_EQ_ADAPTIVE =
struct


datatype u = datatype ListAdaptive.u
datatype t = datatype ListAdaptive.t
datatype alist = datatype ListAdaptive.alist


type 'a key =  'a Key.t
type 'a eq  = 'a Eq.t

val eqList = AdaptiveMeta.Box.eq
val eqCell = fn eq => 
	     fn (c1,c2) =>
                Eq.sum (Eq.unit, Eq.tuple2 (eq, eqList))
                       (case c1 of NIL => Sum.INL () | CONS (h,t) => Sum.INR (h,t),
			case c2 of NIL => Sum.INL () | CONS (h,t) => Sum.INR (h,t))
val eqListPair = fn x => Eq.tuple2 (eqList, eqList) x
val eqCellPair = fn eq => fn x => Eq.tuple2 (eqCell eq, eqCell eq) x

val hashList = AdaptiveMeta.Box.hash
val hashCell = fn hash => 
	       fn c =>
		  Hash.sum (Hash.unit, Hash.tuple2 (hash, hashList))
			   (case c of NIL => Sum.INL () | CONS (h,t) => Sum.INR (h,t))
val keyList = {eq = eqList, hash = hashList}
val keyCell = fn {eq,hash} => {eq = eqCell eq, hash = hashCell hash}
val eq = eqList
val hash = hashList
val key = keyList


open Adaptive

type 'a putTh = 'a u Adaptive.putTh
type 'a put = 'a u Adaptive.put
val putThEq = fn eq => afn th => Adaptive.putThEq (eqCell eq) $ th
val putEq = fn eq => afn x => Adaptive.putEq (eqCell eq) $ x

structure PutMemoTable =
struct
  type ('key,'a) putMTh = ('key,'a u) Adaptive.PutMemoTable.putMTh
  type ('key,'a) putM = ('key,'a u) Adaptive.PutMemoTable.putM
  type ('key,'a) t = ('key,'a u) Adaptive.PutMemoTable.t
  val mkPut = afn (key,eq) => Adaptive.PutMemoTable.mkPutEq $ (key, eqCell eq)
end

val fromList = fn eqElem => afn l =>
   let
      afun fl l =
         case l of 
            [] => putEq eqElem $ NIL
          | h::t => (putEq eqElem $ (CONS (h, fl $ t)))
   in
      fl $ l
   end

fun lengthLessThan n = afn l =>
   let
      afun len (m, l) =
         if m >= n
            then false
         else 
            case get $ l of
               NIL => true
             | CONS (_, t) => len $ (m + 1, t)

      val len = 
         memoContEq (Box.eq, afn l =>
         Adaptive.putThEq (op =) $ (afn () => len $ (0, l)))
   in
      len $ l
   end

type 'a append = 'a t * 'a t -$> 'a t
fun append keyElem = afn (l,r) =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (keyElem, #eq keyElem))
      val append =
         memoFixEq (keyList, eqList, fn append => afn l =>
         case get $ l of 
            NIL => r
          | CONS (h,t) => (putM $ (h, CONS (h, append $ t))))
   in
      append $ l
   end

type 'a rev = 'a t -$> 'a t
fun rev keyElem = afn l =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val revAcc =
         memoFixEq (Key.tuple2 (keyList, keyList), eqList, fn revAcc => afn (l, acc) =>
         case get $ l of
            NIL => acc
          | CONS (hd, tl) => revAcc $ (tl, putM $ (SOME hd, CONS (hd, acc))))
   in
      revAcc $ (l, putM $ (NONE, NIL))
   end

type ('a,'b) map = ('a -> 'b) -> 'a t -$> 'b t

(* key head's cell *)
fun map (keyIElem,eqOElem) f = afn l =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyIElem, eqOElem))
      val map =
         memoFixEq (keyList, eqList, fn map => afn l =>
         case get $ l of
            NIL => putM $ (NONE, NIL)
          | CONS (hd, tl) => putM $ (SOME hd, CONS (f hd, map $ tl)))
   in
      map $ l
   end
(* key tail's cell *)
local
fun map (keyIElem,eqOElem) f = afn l =>
   let
      val eqOElemCell = eqCell eqOElem
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyIElem, eqOElem))
      val map =
         memoFixEq (keyList, eqOElemCell, fn map => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => CONS (f hd, putM $ (SOME hd, map $ tl)))
   in
      putM $ (NONE, map $ l)
   end
in end

local
fun map_memo (keyIElem,eqOElem) f = afn l =>
   let
      val eqOElemCell = eqCell eqOElem
      val map =
         memoFixEq (keyList, eqOElemCell, fn map => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => CONS (f hd, putThEq eqOElem $ (afn () => map $ tl)))
   in
      putThEq eqOElem $ (afn () => map $ l)
   end
in end

type ('a,'b) map_a = ('a -$> 'b) -> 'a t -$> 'b t
(* key head's cell *)
fun map_a (keyIElem,eqOElem) f = afn l =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyIElem, eqOElem))
      val map =
         memoFixEq (keyList, eqList, fn map => afn l =>
         case get $ l of
            NIL => putM $ (NONE, NIL)
          | CONS (hd, tl) => putM $ (SOME hd, CONS (f $ hd, map $ tl)))
   in
      map $ l
   end
(* key tail's cell *)
local
fun map_a (keyIElem,eqOElem) f = afn l =>
   let
      val eqOElemCell = eqCell eqOElem
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyIElem, eqOElem))
      val map =
         memoFixEq (keyList, eqOElemCell, fn map => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => CONS (f $ hd, putM $ (SOME hd, map $ tl)))
   in
      putM $ (NONE, map $ l)
   end
in end
(* no key, rely on memoization *)
local
fun map_a (keyIElem,eqOElem) f = afn l =>
   let
      val eqOElemCell = eqCell eqOElem
      val map =
         memoFixEq (keyList, eqOElemCell, fn map => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => CONS (f $ hd, putThEq eqOElem $ (afn () => map $ tl)))
   in
      putThEq eqOElem $ (afn () => map $ l)
   end
in end

type 'a filter = ('a -> bool) -> 'a t -$> 'a t
fun filter keyElem p = afn l =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val filter =
         memoFixEq (keyList, eqCell (#eq keyElem), fn filter => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => if p hd then 
                                CONS (hd, putM $ (SOME hd, filter $ tl))
                             else 
                                filter $ tl)
   in
      putM $ (NONE, filter $ l)
   end

type 'a filter_a = ('a -$> bool) -> 'a t -$> 'a t
fun filter_a keyElem p = afn l =>
   let
      val putM = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val filter =
         memoFixEq (keyList, eqCell (#eq keyElem), fn filter => afn l =>
         case get $ l of
            NIL => NIL
          | CONS (hd, tl) => if p $ hd then 
                                CONS (hd, putM $ (SOME hd, filter $ tl))
                             else 
                                filter $ tl)
   in
      putM $ (NONE, filter $ l)
   end


type 'a filter2 = ('a -> bool) * ('a -> bool) -> 'a t -$> 'a t * 'a t
fun filter2 keyElem (p1, p2) = afn l =>
   let
      val putM1 = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val putM2 = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val filter2 =
         memoFixEq (keyList, eqListPair, fn filter2 => afn l =>
         case get $ l of
            NIL => (putM1 $ (NONE, NIL), putM2 $ (NONE, NIL))
          | CONS (hd, tl) => let
                                val (tl1', tl2') = filter2 $ tl
                                val l1' =
                                   if p1 hd then 
                                      putM1 $ (SOME hd, CONS (hd, tl1'))
                                   else 
                                      tl1'
                                val l2' =
                                   if p2 hd
                                      then 
                                         putM2 $ (SOME hd, CONS (hd, tl2'))
                                   else 
                                      tl2'
                             in
                                (l1', l2')
                             end)
   in
      filter2 $ l
   end

type 'a filter2_a = ('a -$> bool) * ('a -$> bool) -> 'a t -$> 'a t * 'a t
fun filter2_a keyElem (p1, p2) = afn l =>
   let
      val putM1 = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val putM2 = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
      val filter2 =
         memoFixEq (keyList, eqListPair, fn filter2 => afn l =>
         case get $ l of
            NIL => (putM1 $ (NONE, NIL), putM2 $ (NONE, NIL))
          | CONS (hd, tl) => let
                                val (tl1', tl2') = filter2 $ tl
                                val l1' =
                                   if p1 $ hd then 
                                      putM1 $ (SOME hd, CONS (hd, tl1'))
                                   else 
                                      tl1'
                                val l2' =
                                   if p2 $ hd
                                      then 
                                         putM2 $ (SOME hd, CONS (hd, tl2'))
                                   else 
                                      tl2'
                             in
                                (l1', l2')
                             end)
   in
      filter2 $ l
   end


type 'a partition = ('a -> bool) -> 'a t -$> 'a t * 'a t
fun partition_aux (putM, keyElem : 'a Key.t) p = afn l =>
   let
      val eqElem = #eq keyElem
      val partition =
         memoFixEq (keyList, eqCellPair eqElem, fn partition => afn l =>
         case get $ l of
            NIL => (NIL, NIL)
          | CONS (hd, tl) => let
                                val (T,F) = partition $ tl
                             in
                                if p hd then 
                                   (CONS (hd, putM $ (Sum.INL hd, T)), F)
                                else 
                                   (T, CONS (hd, putM $ (Sum.INL hd, F)))
                             end)
      val (T,F) = partition $ l
   in
      (putM $ (Sum.INR true, T),
       putM $ (Sum.INR false, F))
   end
fun partition keyElem p = afn l =>
   let
      val eqElem = #eq keyElem
      val putM = #putM (PutMemoTable.mkPut $ (Key.sum (keyElem, Key.bool), eqElem))
   in
      partition_aux (putM, keyElem) p $ l
   end

afun mk_partition keyElem =
   let
      val eqElem = #eq keyElem
      val keyPred : ('a -> bool) Key.t = Key.default
      val putM = 
         #putM (PutMemoTable.mkPut $ (Key.tuple2 (keyPred, 
                                                       Key.sum (keyElem, 
                                                                Key.bool)), 
                                           eqElem))
   in
      fn p => 
      partition_aux (afn (k,x) => putM $ ((p, k), x), keyElem) p
   end


type 'a partition_a = ('a -$> bool) -> 'a t -$> 'a t * 'a t
fun partition_a_aux (putM, keyElem : 'a Key.t) p = afn l =>
   let
      val eqElem = #eq keyElem
      val partition =
         memoFixEq (keyList, eqCellPair eqElem, fn partition => afn l =>
         case get $ l of
            NIL => (NIL, NIL)
          | CONS (hd, tl) => let
                                val (T,F) = partition $ tl
                             in
                                if p $ hd then 
                                   (CONS (hd, putM $ (Sum.INL hd, T)), F)
                                else 
                                   (T, CONS (hd, putM $ (Sum.INL hd, F)))
                             end)
      val (T,F) = partition $ l
   in
      (putM $ (Sum.INR true, T),
       putM $ (Sum.INR false, F))
   end


fun partition_a keyElem p = afn l =>
   let
      val eqElem = #eq keyElem
      val putM = #putM (PutMemoTable.mkPut $ (Key.sum (keyElem, Key.bool), eqElem))
   in
      partition_a_aux (putM, keyElem) p $ l
   end

afun mk_partition_a keyElem =
   let
      val eqElem = #eq keyElem
      val keyPred : ('a -$> bool) Key.t = Key.default
      val putM = 
         #putM (PutMemoTable.mkPut $ (Key.tuple2 (keyPred, 
                                                       Key.sum (keyElem, 
                                                                Key.bool)), 
                                           eqElem))
   in
      fn p => 
      partition_a_aux (afn (k,x) => putM $ ((p, k), x), keyElem) p
   end

type 'a mkSplit = unit -> 'a -> bool
type 'a reduce = 'a mkSplit -> ('a * 'a -> 'a) -> 'a -> 'a t -$> 'a AdaptiveMeta.Box.t
type 'a combine = 'a mkSplit -> ('a * 'a -> 'a) -> 'a t -$> 'a AdaptiveMeta.Box.t

fun gen_fold_a_aux keyElem mkRandSplit f fNil fSing = afn l =>
   let
      afun halfList (l, d) =
         let
           val putM = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
           val randSplit = mkRandSplit ()
               
           afun foldRun (v,l) =
               case get $ l of
                 NIL => (v, l)
               | CONS (h, t) =>
                 let
                   val vh = f $ (v, h)
                 in
                   if randSplit h
                   then (vh, t)
                   else foldRun $ (vh, t)
                 end

	   val half =
               memoFixEq (keyList, eqCell (#eq keyElem), fn half => afn l =>
               case get $ l of
                  NIL => NIL 
                | CONS (h, t) => 
                     let
                        val (v, tt) = foldRun $ (h,t)
                        val ttt = half $ tt
                     in
		       CONS(v, putM $ (SOME h, ttt))
                     end)
	 in
           putM $ (NONE, half $ l)
	 end

      afun combine (l, d) = 
         if get $ ((lengthLessThan 2) $ l) 
            then case get $ l of
                    NIL => fNil ()
                  | CONS(h,_) => fSing h
	 else combine $ (halfList $ (l, d), d + 1)

      val combine = 
         memoContEq (Box.eq, afn l =>  
         Adaptive.putThEq (#eq keyElem) $ (afn () => combine $ (l, 0)))
   in
      combine $ l
   end

fun gen_fold_aux keyElem mkRandSplit f fNil fSing = afn l =>
   let
      afun halfList (l, d) =
         let
           val putM = #putM (PutMemoTable.mkPut $ (Key.option keyElem, #eq keyElem))
           val randSplit = mkRandSplit ()
               
           afun foldRun (v,l) =
               case get $ l of
                 NIL => (v, l)
               | CONS (h, t) =>
                 let
                   val vh = f (v, h)
                 in
                   if randSplit h
                   then (vh, t)
                   else foldRun $ (vh, t)
                 end

	   val half =
               memoFixEq (keyList, eqCell (#eq keyElem), fn half => afn l =>
               case get $ l of
                  NIL => NIL 
                | CONS (h, t) => 
                     let
                        val (v, tt) = foldRun $ (h,t)
                        val ttt = half $ tt
                     in
		       CONS(v, putM $ (SOME h, ttt))
                     end)
	 in
           putM $ (NONE, half $ l)
	 end

      afun combine (l, d) = 
         if get $ ((lengthLessThan 2) $ l) 
            then case get $ l of
                    NIL => fNil ()
                  | CONS(h,_) => fSing h
	 else combine $ (halfList $ (l, d), d + 1)

      val combine = 
         memoContEq (Box.eq, afn l =>  
         Adaptive.putThEq (#eq keyElem) $ (afn () => combine $ (l, 0)))
   in
      combine $ l
   end

fun reduce keyElem mkRandSplit f b = afn l =>
   gen_fold_aux keyElem mkRandSplit f (fn () => b) (fn x => f (b, x)) $ l

fun combine keyElem mkRandSplit f = afn l =>
   gen_fold_aux keyElem mkRandSplit f (fn () => raise Empty) (fn x => x) $ l

type 'a combine_a = 'a mkSplit -> ('a * 'a -$> 'a) -> 'a t -$> 'a AdaptiveMeta.Box.t

fun combine_a keyElem mkRandSplit f = afn l =>
   gen_fold_a_aux keyElem mkRandSplit f (fn () => raise Empty) (fn x => x) $ l

end
