(**********************************************************************
 ** POINT_TWO_D.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Umut Acar, Kanat Tangwongsan, Jorge Vittes
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The signature for a point in 2 dimensions.
 **********************************************************************)

signature POINT_TWO_D = 
 sig
   type t(* = Polynomial.t * Polynomial.t*)

   (* Build a point from (x,y) *)
   val new: (real*real) -> t
   (* Build a moving points from initial positions and velocities *)
   val newMoving: ((real*real)*(real*real)) -> t

   val xsub: t * t -> Polynomial.t
   val ysub: t * t -> Polynomial.t
   val xysub: t * t -> Polynomial.t * Polynomial.t

   val cross: (Polynomial.t  * Polynomial.t) * (Polynomial.t  * Polynomial.t) ->  Polynomial.t

   (* XXX same? *)
   (* (p1-p) x (p2-p) *)
   val crossL: {p1:t, p2:t} * t -> Polynomial.t
   (* (p-p1) x (p-p2) *)
   val crossR: t * {p1:t, p2:t} -> Polynomial.t

   (* The current location of the point *)
   val toPolyLoc: t -> Polynomial.t*Polynomial.t
   val fromPolyLoc: Polynomial.t*Polynomial.t -> t

   (* Static position i.e. position at time 0 *)
   val locAtTime0: t -> real*real
   val locAtTime: real -> t -> real*real
   
   val atTime: t * real -> t

   val toString: t -> string
   val eq: t Eq.t
   val hash: t Hash.t
   val key: t Key.t
 end
