(**********************************************************************
 ** PointTwoD.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Umut Acar, Kanat Tangwongsan, Jorge Vittes
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: A 2-d point.
 **********************************************************************)
structure PointTwoD :> POINT_TWO_D = 
 struct
   structure Poly = Polynomial

   type t = Poly.t*Poly.t (* poly in t of x and y *)

   fun new (x,y) = (Poly.fromRealList[x],Poly.fromRealList[y])
   fun newMoving ((x,y), (vx,vy)) = (Poly.fromRealList[x,vx],Poly.fromRealList[y,vy])

   fun toPolyLoc p = p
   fun fromPolyLoc p = p

   fun fromLoc (x,y) = newMoving ((x,y), (0.0,0.0))

   val xsub =
    fn ((x1,_), (x2,_)) =>
       Poly.sub (x1,x2)
   val ysub =
    fn ((_,y1), (_,y2)) =>
       Poly.sub (y1,y2)
   val xysub =
    fn ps =>
       (xsub ps, ysub ps)

   val cross =
    fn ((x1,y1), (x2,y2)) =>
       Poly.sub(Poly.mult(x1, y2), Poly.mult(x2, y1))

   val crossL =
    fn ({p1,p2},p) =>
       cross (xysub (p1,p), xysub (p2,p))
   val crossR =
    fn (p,{p1,p2}) =>
       cross (xysub (p,p1), xysub (p,p2))

   fun locAtTime0 (px, py) = (Poly.eval0 px, Poly.eval0 py)
   fun locAtTime time (px, py) = (Poly.eval (px, time), Poly.eval (py, time))
   fun atTime (p,t) = (fromLoc (locAtTime t p))
			   
   fun toString (xs,ys) =
       let
	   val (x::_,y::_) = (Poly.toRealList xs, Poly.toRealList ys)
       in
	   "(" ^Real.toString(x) ^ "," ^ Real.toString(y) ^ ")"
       end
  

   val eq = Eq.tuple2 (Poly.eq, Poly.eq)

   val hash = Hash.tuple2 (Poly.hash, Poly.hash)
   val key = Key.tuple2 (Poly.key, Poly.key)
 end
