(***************************************************************************
 ** Certificate.sml
 ** sml
 **
 ** Umut Acar, Kanat Tangwongsan and Jorge Vittes
 **
 ** Certificate creation
 ****************************************************************************)
structure Certify : CERTIFICATE =
struct

type t = bool AdaptiveMeta.Box.t
type certificate = t

exception Error
exception RestartSimulator

structure Simulator = SimulateMotion

fun boolComp(a,b) = (a=b)

fun splitAtTime (sign,time, cutoff,roots) = 
  let
    fun get (s,l) = 
      case l of 
        nil => (s,nil)
      | h::t =>
        if (Real.>(h,cutoff)) then 
	  (s,l)
        else  if (Real.> (h,time))  then
	  (Simulator.incNFlips (); get(not s,t))
	else
	  get(not s,t)
  in
    get(sign,roots)
  end


fun mkCertificate' (roots, signneginf) =
let
    val t = Simulator.getTime()

    val (sign,rest) = splitAtTime (signneginf, t, Simulator.cutOffTime (), roots)
    val _ = case rest of
	      h::_ => if (Simulator.getTime () > 0.0) then
			  (
(*  Umut: this is with the old Epsilon-Delta Scheme.

			  if (Real.>(h, Simulator.cutOffTime ()) andalso 
			      Real.< (h,Simulator.cutOffTimeUpper())) then 
*)
(** Umut; What follows is the new, Epsilon-Delta Scheme. **)
			  if (Real.>(h, Simulator.getTime ()) andalso 
			      Real.< (h,Simulator.cutOffTimeUpper())) then 

			      ( print ("Die Gracefully at time = " ^ Real.toString t ^ 
				       " h = " ^ Real.toString h ^ " \n"); 
				print ("difference = " ^ Real.toString (Real.-(h,t)) ^"\n"); 
				raise RestartSimulator  
			      )
			      else ())
		      else (
			  if (Real.< (h, Simulator.EPSILON)) then
      			      ( print ("ZERO: Die Gracefully at time = " ^ Real.toString t ^ 
				       " h = " ^ Real.toString h ^ " \n"); 
				print ("difference = " ^ Real.toString (Real.-(h,t)) ^"\n"); 
				raise RestartSimulator )
			  else ())

	    | _ => ()

    val c = AdaptiveMeta.newEq (Eq.default, sign)
    val _ = case rest of
	      nil => ()
	    | h::r =>( Simulator.insert(c,h,r); ())
in c
end

fun mkCertificate (poly:real list)= 
  let
    val poly' = poly
    val roots = (PSolver.allToggleSign poly')
    val signneginf = Polynomial.evalneginf poly'
	
    val	c = mkCertificate' (roots, signneginf)
  in
    c
  end
  
end

