(**********************************************************************
 ** BuildOrdListInOut.sml
 ** 
 **********************************************************************)

functor BuildOrdListInOut(structure Elt : ELT
                          structure NAElt : ELT
		          val convertElt: Elt.t -> NAElt.t) =
struct
  fun eltListToString l = 
    let val str = List.foldr 
		    (fn (h,str) =>  (Elt.toString h) ^ ", " ^ str) 
		    "nil"
		    l
    in
      "[" ^  str ^ "]"
    end

  fun naeltListToString l = 
    let val str = List.foldr 
		    (fn (h,str) =>  (NAElt.toString h) ^ ", " ^ str) 
		    "nil"
		    l
    in
      "[" ^  str ^ "]"
    end

   structure Adaptive = 
      struct 
         type t = Elt.t list
         val toString = eltListToString
      end

   structure NonAdaptive = 
      struct
         type t = NAElt.t list
	 val toString = naeltListToString
      end

   structure Converter =
      struct
         type a_t = Adaptive.t
         type na_t = NonAdaptive.t
         fun convert (l : a_t) : na_t = 
           case l of
	     nil =>  nil
	   | h::t => (convertElt h)::(convert t)
      end

   structure Checker =
      struct
         type a_t = Adaptive.t
         type na_t = NonAdaptive.t
         fun check (l : a_t, l' : na_t) =
            case (l, l') of
               (nil, nil) => true
             | (x::t, x'::t') => NAElt.eq (convertElt x, x') andalso check (t, t')
             | _ => false
      end
end
